searchData={"items":[{"type":"module","title":"Brama","doc":"Brama is a library for managing connections to external dependencies.\n\nIt provides features for:\n- Connection monitoring\n- Circuit breaking\n- Self-healing\n- Status notifications","ref":"Brama.html"},{"type":"module","title":"Connection Management - Brama","doc":"Brama allows you to register connections to external services and track their status:\n\n```elixir\n# Register a connection\nBrama.register(\"payment_api\")\n\n# Check if a connection is available\nBrama.available?(\"payment_api\")\n\n# Report success/failure\nBrama.success(\"payment_api\")\nBrama.failure(\"payment_api\", reason: \"Timeout\")\n```","ref":"Brama.html#module-connection-management"},{"type":"module","title":"Circuit Breaking - Brama","doc":"Brama implements the circuit breaker pattern to prevent cascading failures:\n\n```elixir\n# Circuit opens automatically after consecutive failures\nBrama.failure(\"payment_api\", reason: \"Timeout\") # After max_attempts, circuit opens\n\n# Manual circuit control\nBrama.open_circuit!(\"payment_api\", reason: \"Maintenance\")\nBrama.close_circuit!(\"payment_api\")\nBrama.reset_circuit!(\"payment_api\")\n```","ref":"Brama.html#module-circuit-breaking"},{"type":"module","title":"Configuration - Brama","doc":"Brama supports both global and per-connection configuration:\n\n```elixir\n# Update global settings\nBrama.configure(max_attempts: 15, expiry: 120_000)\n\n# Update specific connection\nBrama.configure(\"payment_api\", max_attempts: 5, expiry: 30_000)\n\n# Register with progressive backoff\nBrama.register(\"flaky_service\",\n  expiry_strategy: :progressive,\n  initial_expiry: 10_000,\n  max_expiry: 300_000,\n  backoff_factor: 2.0\n)\n```","ref":"Brama.html#module-configuration"},{"type":"module","title":"Event Subscription - Brama","doc":"Subscribe to events to get notified of state changes:\n\n```elixir\n# Subscribe to all events for a specific connection\nBrama.subscribe(connection: \"payment_api\")\n\n# Subscribe to specific event types\nBrama.subscribe(events: [:state_change, :failure])\n\n# Subscribe with a filter function\nBrama.subscribe(filter: fn event -> event.connection == \"payment_api\" end)\n```","ref":"Brama.html#module-event-subscription"},{"type":"function","title":"Brama.available?/2","doc":"Checks if a connection is available (circuit is closed or half-open).","ref":"Brama.html#available?/2"},{"type":"function","title":"Parameters - Brama.available?/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match","ref":"Brama.html#available?/2-parameters"},{"type":"function","title":"Examples - Brama.available?/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.available?(\"payment_api\")\ntrue\n```","ref":"Brama.html#available?/2-examples"},{"type":"function","title":"Brama.close_circuit!/2","doc":"Manually closes the circuit for a connection.","ref":"Brama.html#close_circuit!/2"},{"type":"function","title":"Parameters - Brama.close_circuit!/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match","ref":"Brama.html#close_circuit!/2-parameters"},{"type":"function","title":"Examples - Brama.close_circuit!/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.close_circuit!(\"payment_api\")\n:ok\n```","ref":"Brama.html#close_circuit!/2-examples"},{"type":"function","title":"Brama.configure/2","doc":"Updates the configuration for a specific connection or globally.","ref":"Brama.html#configure/2"},{"type":"function","title":"Parameters - Brama.configure/2","doc":"- `identifier` - Optional connection identifier. If not provided, updates global settings\n- `opts` - Configuration options to update:\n  - `:max_attempts` - Maximum number of failures before opening circuit\n  - `:expiry` - Time in milliseconds before open circuit transitions to half-open\n  - `:expiry_strategy` - Either `:fixed` or `:progressive` for backoff\n  - `:initial_expiry` - Initial expiry time for progressive backoff\n  - `:max_expiry` - Maximum expiry time for progressive backoff\n  - `:backoff_factor` - Factor for exponential backoff","ref":"Brama.html#configure/2-parameters"},{"type":"function","title":"Examples - Brama.configure/2","doc":"```elixir\n# Update global settings\nBrama.configure(max_attempts: 15, expiry: 120_000)\n\n# Update specific connection\nBrama.configure(\"payment_api\", max_attempts: 5, expiry: 30_000)\n\n# Configure progressive backoff\nBrama.configure(\"flaky_service\",\n  expiry_strategy: :progressive,\n  initial_expiry: 10_000,\n  max_expiry: 300_000,\n  backoff_factor: 2.0\n)\n```","ref":"Brama.html#configure/2-examples"},{"type":"function","title":"Brama.failure/2","doc":"Reports a failed connection attempt.","ref":"Brama.html#failure/2"},{"type":"function","title":"Parameters - Brama.failure/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match\n  - `:reason`: Reason for the failure\n  - `:metadata`: Additional metadata about the failure","ref":"Brama.html#failure/2-parameters"},{"type":"function","title":"Examples - Brama.failure/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.failure(\"payment_api\", reason: \"Timeout\")\n:ok\n```","ref":"Brama.html#failure/2-examples"},{"type":"function","title":"Brama.open_circuit!/2","doc":"Manually opens the circuit for a connection.","ref":"Brama.html#open_circuit!/2"},{"type":"function","title":"Parameters - Brama.open_circuit!/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match\n  - `:reason`: Reason for opening the circuit\n  - `:expiry`: Custom expiry time in milliseconds","ref":"Brama.html#open_circuit!/2-parameters"},{"type":"function","title":"Examples - Brama.open_circuit!/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.open_circuit!(\"payment_api\", reason: \"Maintenance\")\n:ok\n```","ref":"Brama.html#open_circuit!/2-examples"},{"type":"function","title":"Brama.register/2","doc":"Registers a new connection with Brama.","ref":"Brama.html#register/2"},{"type":"function","title":"Parameters - Brama.register/2","doc":"- `identifier`: A unique identifier for the connection\n- `opts`: Options for the connection\n  - `:scope`: Optional scope for grouping connections\n  - `:max_attempts`: Maximum number of failures before opening the circuit (default: 5)\n  - `:expiry`: Time in milliseconds after which an open circuit transitions to half-open (default: 60000)\n  - `:metadata`: Additional metadata to store with the connection","ref":"Brama.html#register/2-parameters"},{"type":"function","title":"Examples - Brama.register/2","doc":"```elixir\niex> {:ok, result} = Brama.register(\"payment_api\")\niex> Map.take(result, [:identifier, :state, :failure_count])\n%{identifier: \"payment_api\", state: :closed, failure_count: 0}\n\niex> {:ok, result} = Brama.register(\"invoice_api\", scope: \"billing\", max_attempts: 5)\niex> Map.take(result, [:identifier, :scope, :state, :failure_count])\n%{identifier: \"invoice_api\", scope: \"billing\", state: :closed, failure_count: 0}\n```","ref":"Brama.html#register/2-examples"},{"type":"function","title":"Brama.reset_circuit!/2","doc":"Resets the circuit for a connection (closes it and resets failure count).","ref":"Brama.html#reset_circuit!/2"},{"type":"function","title":"Parameters - Brama.reset_circuit!/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match","ref":"Brama.html#reset_circuit!/2-parameters"},{"type":"function","title":"Examples - Brama.reset_circuit!/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.reset_circuit!(\"payment_api\")\n:ok\n```","ref":"Brama.html#reset_circuit!/2-examples"},{"type":"function","title":"Brama.status/2","doc":"Gets the current status of a connection.","ref":"Brama.html#status/2"},{"type":"function","title":"Parameters - Brama.status/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match","ref":"Brama.html#status/2-parameters"},{"type":"function","title":"Examples - Brama.status/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> {:ok, result} = Brama.status(\"payment_api\")\niex> Map.take(result, [:state, :failure_count])\n%{state: :closed, failure_count: 0}\n```","ref":"Brama.html#status/2-examples"},{"type":"function","title":"Brama.subscribe/1","doc":"Subscribes to connection events.","ref":"Brama.html#subscribe/1"},{"type":"function","title":"Parameters - Brama.subscribe/1","doc":"- `opts`: Subscription options (at least one must be provided)\n  - `:connection`: Filter events by connection identifier\n  - `:scope`: Filter events by connection scope\n  - `:events`: List of event types to subscribe to\n  - `:filter`: Custom filter function","ref":"Brama.html#subscribe/1-parameters"},{"type":"function","title":"Examples - Brama.subscribe/1","doc":"```elixir\niex> Brama.subscribe(events: [:state_change])\n{:ok, pid}\n```","ref":"Brama.html#subscribe/1-examples"},{"type":"function","title":"Brama.success/2","doc":"Reports a successful connection attempt.","ref":"Brama.html#success/2"},{"type":"function","title":"Parameters - Brama.success/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match\n  - `:metadata`: Additional metadata about the success","ref":"Brama.html#success/2-parameters"},{"type":"function","title":"Examples - Brama.success/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.success(\"payment_api\")\n:ok\n```","ref":"Brama.html#success/2-examples"},{"type":"function","title":"Brama.unregister/2","doc":"Unregisters a connection from Brama.","ref":"Brama.html#unregister/2"},{"type":"function","title":"Parameters - Brama.unregister/2","doc":"- `identifier`: The connection identifier\n- `opts`: Options\n  - `:scope`: Optional scope to match","ref":"Brama.html#unregister/2-parameters"},{"type":"function","title":"Examples - Brama.unregister/2","doc":"```elixir\niex> Brama.register(\"payment_api\")\niex> Brama.unregister(\"payment_api\")\n:ok\n```","ref":"Brama.html#unregister/2-examples"},{"type":"function","title":"Brama.unsubscribe/1","doc":"Unsubscribes from connection events.","ref":"Brama.html#unsubscribe/1"},{"type":"function","title":"Parameters - Brama.unsubscribe/1","doc":"- `subscription_id`: The subscription ID returned from subscribe/1","ref":"Brama.html#unsubscribe/1-parameters"},{"type":"function","title":"Examples - Brama.unsubscribe/1","doc":"```elixir\niex> {:ok, subscription} = Brama.subscribe(events: [:state_change])\niex> Brama.unsubscribe(subscription)\n:ok\n```","ref":"Brama.html#unsubscribe/1-examples"},{"type":"module","title":"Brama.ConnectionManager","doc":"Manages connection states and implements circuit breaking logic.\n\nThis module is responsible for:\n- Tracking connection states (closed, open, half-open)\n- Implementing circuit breaking logic\n- Managing state transitions\n- Cleaning up expired connections","ref":"Brama.ConnectionManager.html"},{"type":"function","title":"Brama.ConnectionManager.available?/2","doc":"Checks if a connection is available (circuit is closed or half-open).","ref":"Brama.ConnectionManager.html#available?/2"},{"type":"function","title":"Options - Brama.ConnectionManager.available?/2","doc":"* `:scope` - Optional scope to match","ref":"Brama.ConnectionManager.html#available?/2-options"},{"type":"function","title":"Brama.ConnectionManager.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Brama.ConnectionManager.html#child_spec/1"},{"type":"function","title":"Brama.ConnectionManager.close_circuit!/2","doc":"Manually closes the circuit for a connection.","ref":"Brama.ConnectionManager.html#close_circuit!/2"},{"type":"function","title":"Options - Brama.ConnectionManager.close_circuit!/2","doc":"* `:scope` - Optional scope to match","ref":"Brama.ConnectionManager.html#close_circuit!/2-options"},{"type":"function","title":"Brama.ConnectionManager.configure/2","doc":"Updates the configuration for a specific connection or globally.","ref":"Brama.ConnectionManager.html#configure/2"},{"type":"function","title":"Parameters - Brama.ConnectionManager.configure/2","doc":"- `identifier` - Optional connection identifier. If not provided, updates global settings\n- `opts` - Configuration options to update","ref":"Brama.ConnectionManager.html#configure/2-parameters"},{"type":"type","title":"Brama.ConnectionManager.connection_data/0","doc":"","ref":"Brama.ConnectionManager.html#t:connection_data/0"},{"type":"type","title":"Brama.ConnectionManager.connection_id/0","doc":"","ref":"Brama.ConnectionManager.html#t:connection_id/0"},{"type":"type","title":"Brama.ConnectionManager.connection_scope/0","doc":"","ref":"Brama.ConnectionManager.html#t:connection_scope/0"},{"type":"type","title":"Brama.ConnectionManager.connection_state/0","doc":"","ref":"Brama.ConnectionManager.html#t:connection_state/0"},{"type":"function","title":"Brama.ConnectionManager.failure/2","doc":"Reports a failed connection attempt.","ref":"Brama.ConnectionManager.html#failure/2"},{"type":"function","title":"Options - Brama.ConnectionManager.failure/2","doc":"* `:scope` - Optional scope to match\n* `:reason` - Reason for the failure\n* `:metadata` - Additional metadata about the failure","ref":"Brama.ConnectionManager.html#failure/2-options"},{"type":"function","title":"Brama.ConnectionManager.open_circuit!/2","doc":"Manually opens the circuit for a connection.","ref":"Brama.ConnectionManager.html#open_circuit!/2"},{"type":"function","title":"Options - Brama.ConnectionManager.open_circuit!/2","doc":"* `:scope` - Optional scope to match\n* `:reason` - Reason for opening the circuit\n* `:expiry` - Custom expiry time in milliseconds","ref":"Brama.ConnectionManager.html#open_circuit!/2-options"},{"type":"function","title":"Brama.ConnectionManager.register/2","doc":"Registers a new connection.","ref":"Brama.ConnectionManager.html#register/2"},{"type":"function","title":"Options - Brama.ConnectionManager.register/2","doc":"* `:scope` - Optional scope for grouping connections\n* `:max_attempts` - Maximum number of failures before opening circuit\n* `:expiry` - Time in milliseconds after which an open circuit transitions to half-open\n* `:metadata` - Additional metadata to store with the connection","ref":"Brama.ConnectionManager.html#register/2-options"},{"type":"function","title":"Brama.ConnectionManager.reset_circuit!/2","doc":"Resets the circuit for a connection (closes it and resets failure count).","ref":"Brama.ConnectionManager.html#reset_circuit!/2"},{"type":"function","title":"Options - Brama.ConnectionManager.reset_circuit!/2","doc":"* `:scope` - Optional scope to match","ref":"Brama.ConnectionManager.html#reset_circuit!/2-options"},{"type":"function","title":"Brama.ConnectionManager.start_link/1","doc":"Starts the connection manager.","ref":"Brama.ConnectionManager.html#start_link/1"},{"type":"function","title":"Brama.ConnectionManager.status/2","doc":"Gets the current status of a connection.","ref":"Brama.ConnectionManager.html#status/2"},{"type":"function","title":"Options - Brama.ConnectionManager.status/2","doc":"* `:scope` - Optional scope to match","ref":"Brama.ConnectionManager.html#status/2-options"},{"type":"function","title":"Brama.ConnectionManager.success/2","doc":"Reports a successful connection attempt.","ref":"Brama.ConnectionManager.html#success/2"},{"type":"function","title":"Options - Brama.ConnectionManager.success/2","doc":"* `:scope` - Optional scope to match\n* `:metadata` - Additional metadata about the success","ref":"Brama.ConnectionManager.html#success/2-options"},{"type":"function","title":"Brama.ConnectionManager.unregister/2","doc":"Unregisters a connection.","ref":"Brama.ConnectionManager.html#unregister/2"},{"type":"function","title":"Options - Brama.ConnectionManager.unregister/2","doc":"* `:scope` - Optional scope to match","ref":"Brama.ConnectionManager.html#unregister/2-options"},{"type":"module","title":"Brama.Decorator","doc":"Provides macros for decorating functions with circuit breaker functionality to improve system resilience\nand handle failures gracefully.","ref":"Brama.Decorator.html"},{"type":"module","title":"Overview - Brama.Decorator","doc":"The circuit breaker pattern helps prevent cascading failures by monitoring service calls and\nautomatically stopping requests when a service is experiencing issues. It operates in three states:\n- Closed (normal operation)\n- Open (service calls blocked)\n- Half-open (testing if service recovered)","ref":"Brama.Decorator.html#module-overview"},{"type":"module","title":"Configuration - Brama.Decorator","doc":"The decorator accepts the following options:\n* `identifier` - (required) A unique string identifier for the circuit breaker instance\n* `error_handler` - (optional) A function to handle responses and determine success/failure status","ref":"Brama.Decorator.html#module-configuration"},{"type":"module","title":"Error Handling - Brama.Decorator","doc":"The error handler function can return the following statuses:\n* `:success` - Operation completed successfully, keeps circuit closed\n* `:failure` - Operation failed, increments failure count\n* `{:failure, reason}` - Operation failed with specific reason\n* `:ignore` - Operation result should not affect circuit breaker state\n\nDefault error handler behavior:\n```elixir\ndef default_error_handler(result) do\n  case result do\n    {:ok, _} -> :success\n    {:error, reason} -> {:failure, reason}\n    _ -> :ignore\n  end\nend\n```","ref":"Brama.Decorator.html#module-error-handling"},{"type":"module","title":"Exception Handling - Brama.Decorator","doc":"When an exception occurs:\n1. The exception is logged for debugging\n2. A failure is recorded with the exception message as reason\n3. The original exception is re-raised with preserved stacktrace","ref":"Brama.Decorator.html#module-exception-handling"},{"type":"module","title":"Service Unavailability - Brama.Decorator","doc":"When the circuit is open:\n1. The decorated function is not executed\n2. Returns `{:error, status}` where status contains circuit breaker state\n3. Prevents overwhelming failing services with requests","ref":"Brama.Decorator.html#module-service-unavailability"},{"type":"module","title":"Examples - Brama.Decorator","doc":"Basic usage with default error handling:\n```elixir\ndefmodule PaymentService do\n  use Brama.Decorator\n\n  @decorate circuit_breaker(identifier: \"payment_api\")\n  def process_payment(payment) do\n    PaymentAPI.process(payment)\n  end\nend\n```\n\nCustom error handling:\n```elixir\ndefmodule PaymentService do\n  use Brama.Decorator\n\n  @decorate circuit_breaker(\n    identifier: \"refund_api\",\n    error_handler: fn\n      {:ok, _} -> :success\n      {:error, :invalid_amount} -> {:failure, :validation_error}\n      {:error, :network_timeout} -> {:failure, :service_unavailable}\n      _ -> :ignore\n    end\n  )\n  def process_refund(refund) do\n    RefundAPI.process(refund)\n  end\nend\n```","ref":"Brama.Decorator.html#module-examples"},{"type":"module","title":"Return Values - Brama.Decorator","doc":"The decorated function preserves the original return value when the circuit is closed.\nWhen the circuit is open, it returns:\n```elixir\n{:error, status} # where status contains circuit breaker state information\n```","ref":"Brama.Decorator.html#module-return-values"},{"type":"macro","title":"Brama.Decorator.circuit_breaker/1","doc":"","ref":"Brama.Decorator.html#circuit_breaker/1"},{"type":"function","title":"Brama.Decorator.circuit_breaker/3","doc":"Decorator function that wraps a function with circuit breaker functionality.","ref":"Brama.Decorator.html#circuit_breaker/3"},{"type":"function","title":"Options - Brama.Decorator.circuit_breaker/3","doc":"- `identifier` - Connection identifier (required)\n- `error_handler` - Custom function to handle errors and determine success/failure","ref":"Brama.Decorator.html#circuit_breaker/3-options"},{"type":"function","title":"Examples - Brama.Decorator.circuit_breaker/3","doc":"```elixir\n@decorate circuit_breaker(identifier: \"payment_api\")\ndef process_payment(payment) do\n  # This function will be wrapped with circuit breaker logic\n  PaymentAPI.process(payment)\nend\n```","ref":"Brama.Decorator.html#circuit_breaker/3-examples"},{"type":"function","title":"Brama.Decorator.default_error_handler/1","doc":"","ref":"Brama.Decorator.html#default_error_handler/1"},{"type":"function","title":"Brama.Decorator.ensure_registered/1","doc":"","ref":"Brama.Decorator.html#ensure_registered/1"},{"type":"module","title":"Brama.EventDispatcher","doc":"Handles event dispatching and subscription management for Brama.\n\nThis module is responsible for:\n- Maintaining a registry of event subscribers\n- Filtering events based on subscriber preferences\n- Dispatching events to interested subscribers","ref":"Brama.EventDispatcher.html"},{"type":"function","title":"Brama.EventDispatcher.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Brama.EventDispatcher.html#child_spec/1"},{"type":"function","title":"Brama.EventDispatcher.dispatch/4","doc":"Dispatches an event to all matching subscribers.\n\nThis function constructs an event map and notifies all subscribers\nwhose filters match the event.","ref":"Brama.EventDispatcher.html#dispatch/4"},{"type":"function","title":"Brama.EventDispatcher.start_link/1","doc":"Starts the event dispatcher.","ref":"Brama.EventDispatcher.html#start_link/1"},{"type":"function","title":"Brama.EventDispatcher.subscribe/1","doc":"Subscribes to Brama events.","ref":"Brama.EventDispatcher.html#subscribe/1"},{"type":"function","title":"Options - Brama.EventDispatcher.subscribe/1","doc":"* `:events` - List of event types to subscribe to (default: all)\n* `:connection` - Connection identifier to filter by (default: all)\n* `:scope` - Connection scope to filter by (default: all)","ref":"Brama.EventDispatcher.html#subscribe/1-options"},{"type":"function","title":"Brama.EventDispatcher.unsubscribe/1","doc":"Unsubscribes from Brama events.","ref":"Brama.EventDispatcher.html#unsubscribe/1"},{"type":"module","title":"Brama.TestHelpers","doc":"Provides helper functions for testing applications that use Brama.\n\nThis module is intended to be used in test environments to:\n- Manipulate circuit states directly\n- Simulate failures and recovery","ref":"Brama.TestHelpers.html"},{"type":"module","title":"Usage - Brama.TestHelpers","doc":"```elixir\n# In your test file\nuse ExUnit.Case\nimport Brama.TestHelpers\n\ntest \"circuit opens after failures\" do\n  # Setup\n  Brama.register(\"test_api\")\n\n  # Simulate failures\n  add_failures(\"test_api\", 10)\n\n  # Assert circuit state\n  assert {:ok, %{state: :open}} = Brama.status(\"test_api\")\nend\n```","ref":"Brama.TestHelpers.html#module-usage"},{"type":"function","title":"Brama.TestHelpers.add_failures/3","doc":"Adds a specific number of failures to a connection.","ref":"Brama.TestHelpers.html#add_failures/3"},{"type":"function","title":"Examples - Brama.TestHelpers.add_failures/3","doc":"iex> add_failures(\"payment_api\", 5)\n    :ok","ref":"Brama.TestHelpers.html#add_failures/3-examples"},{"type":"function","title":"Brama.TestHelpers.set_state/3","doc":"Directly sets the state of a circuit.","ref":"Brama.TestHelpers.html#set_state/3"},{"type":"function","title":"Examples - Brama.TestHelpers.set_state/3","doc":"iex> set_state(\"payment_api\", :open)\n    :ok","ref":"Brama.TestHelpers.html#set_state/3-examples"},{"type":"function","title":"Brama.TestHelpers.simulate_failures/3","doc":"Simulates a service temporarily failing.","ref":"Brama.TestHelpers.html#simulate_failures/3"},{"type":"function","title":"Examples - Brama.TestHelpers.simulate_failures/3","doc":"iex> simulate_failures(\"payment_api\", 5)\n    :ok","ref":"Brama.TestHelpers.html#simulate_failures/3-examples"},{"type":"function","title":"Brama.TestHelpers.simulate_recovery/2","doc":"Simulates a service recovering.","ref":"Brama.TestHelpers.html#simulate_recovery/2"},{"type":"function","title":"Examples - Brama.TestHelpers.simulate_recovery/2","doc":"iex> simulate_recovery(\"payment_api\")\n    :ok","ref":"Brama.TestHelpers.html#simulate_recovery/2-examples"},{"type":"extras","title":"Brama","doc":"# Brama\n\nBrama is an Elixir library for reliable connection management with external dependencies. It provides robust tracking of connection statuses to APIs, services, databases, or any external system, with built-in circuit breaking to prevent cascading failures.\n\n> This library is my first attempt at utilizing the [\"stdlib\" approach](https://ghuntley.com/stdlib/) for building software. This method involves creating comprehensive specifications before implementation, building a collection of Cursor rules to guide development, and treating AI as an autonomous agent rather than just a coding assistant. The whole library is going to be written by Cursor Agent with my supervision.","ref":"readme.html"},{"type":"extras","title":"Overview - Brama","doc":"When your application depends on external systems, knowing their availability status becomes critical. Brama serves as a gatekeeper (the name \"Brama\" means \"gate\" in several languages), monitoring your connections and protecting your application from external system failures.","ref":"readme.html#overview"},{"type":"extras","title":"Features - Brama","doc":"- **Connection Monitoring**: Track status of any connection\n- **Circuit Breaking**: Automatically prevent requests to failing systems after a threshold is reached\n- **Self-Healing**: Connections automatically reset after a configurable expiry time\n- **Status Notifications**: Subscribe to connection status change events\n- **Failure Isolation**: Protect your application from cascading failures\n- **Minimal Configuration**: Simple setup with reasonable defaults\n- **Decorator API**: Simple function decorators for automatic circuit breaking\n- **Flexible Expiry Strategies**: Fixed or progressive backoff strategies","ref":"readme.html#features"},{"type":"extras","title":"Installation - Brama","doc":"Add Brama to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:brama, \"~> 0.1.0\"},\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Circuit Breaking Mechanism - Brama","doc":"Brama implements circuit breaking for any connection with these behaviors:\n\n1. Each connection type is tracked separately\n2. After a configurable number of failed attempts (default: 10), the circuit opens\n3. When open, all requests are rejected without attempting the external call\n4. After a configurable time period (default: 1 minute), the circuit transitions to half-open\n5. Connections have immediate status updates","ref":"readme.html#circuit-breaking-mechanism"},{"type":"extras","title":"Configuration Options - Brama","doc":"```elixir\nconfig :brama,\n  max_attempts: 10,           # Attempts before circuit opens\n  cleanup_interval: 10_000,   # Status check interval in ms\n  expiry: 60_000              # Circuit open duration in ms\n```","ref":"readme.html#configuration-options"},{"type":"extras","title":"Basic Usage - Brama","doc":"```elixir\n# Register a connection\nBrama.register(\"payment_api\", type: :http, scope: \"payment_services\")\n\n# Before making an external call\nif Brama.available?(\"payment_api\") do\n  case make_api_call() do\n    {:ok, result} -> \n      Brama.success(\"payment_api\")\n      {:ok, result}\n    {:error, reason} -> \n      Brama.failure(\"payment_api\")\n      {:error, reason}\n  end\nelse\n  # Use fallback or return error\n  {:error, :service_unavailable}\nend\n```","ref":"readme.html#basic-usage"},{"type":"extras","title":"Decorator API - Brama","doc":"For a cleaner integration, use the decorator API:\n\n```elixir\ndefmodule PaymentService do\n  use Brama.Decorator\n\n  @decorate circuit_breaker(identifier: \"payment_api\")\n  def process_payment(payment) do\n    PaymentAPI.process(payment)\n  end\n  \n  @decorate circuit_breaker(\n    identifier: \"refund_api\",\n    error_handler: fn\n      {:ok, _} -> :success\n      {:error, :invalid_amount} -> {:failure, :validation_error}\n      {:error, :network_timeout} -> {:failure, :service_unavailable}\n      _ -> :ignore\n    end\n  )\n  def process_refund(refund) do\n    RefundAPI.process(refund)\n  end\nend\n```","ref":"readme.html#decorator-api"},{"type":"extras","title":"Expiry Strategies - Brama","doc":"Brama supports different strategies for determining how long a circuit remains open:","ref":"readme.html#expiry-strategies"},{"type":"extras","title":"Fixed Expiry (Default) - Brama","doc":"```elixir\nBrama.configure(\"payment_api\",\n  expiry_strategy: :fixed,  # This is the default strategy\n  expiry: 30_000            # 30 seconds\n)\n```","ref":"readme.html#fixed-expiry-default"},{"type":"extras","title":"Progressive Backoff - Brama","doc":"```elixir\nBrama.configure(\"flaky_service\",\n  expiry_strategy: :progressive,\n  initial_expiry: 5_000,     # Start with 5 seconds\n  max_expiry: 300_000,       # Cap at 5 minutes\n  backoff_factor: 2.0        # Double the time with each failure\n)\n```\n\nWith progressive backoff, expiry times would increase with each failure: 5s → 10s → 20s → 40s → etc.","ref":"readme.html#progressive-backoff"},{"type":"extras","title":"Event Notifications - Brama","doc":"Subscribe to connection status changes:\n\n```elixir\n# Subscribe to all state changes\nBrama.subscribe(events: [:state_change])\n\n# Handle events in a GenServer\ndef handle_info({:brama_event, event}, state) do\n  Logger.info(\"Connection #{event.connection} changed to #{event.data.new_state}\")\n  {:noreply, state}\nend\n```","ref":"readme.html#event-notifications"},{"type":"extras","title":"Advanced Usage - Brama","doc":"You can extend Brama for specific needs:\n\n- Create custom monitoring modules for specialized protocols\n- Implement advanced health checking logic\n- Build metrics collection around connection statuses\n- Integrate with monitoring and alerting systems","ref":"readme.html#advanced-usage"},{"type":"extras","title":"Contributing - Brama","doc":"Contributions are welcome! Please feel free to submit a Pull Request.","ref":"readme.html#contributing"},{"type":"extras","title":"License - Brama","doc":"Brama is released under the MIT License.","ref":"readme.html#license"},{"type":"extras","title":"Brama Library Specifications","doc":"# Brama Library Specifications","ref":"specs.html"},{"type":"extras","title":"Overview - Brama Library Specifications","doc":"This document serves as an index for all specification documents related to the Brama library, which provides reliable connection management with external dependencies.\n\nBrama is designed to act as a gatekeeper for your application's interactions with external systems. It monitors connection statuses, implements circuit breaking to prevent cascading failures, and provides self-healing mechanisms and notifications for connection status changes.","ref":"specs.html#overview"},{"type":"extras","title":"Specifications Index - Brama Library Specifications","doc":"| Domain Area | Description | Link |\n|-------------|-------------|------|\n| Architecture | Overall system architecture and design principles | [Architecture](specs/architecture.md) |\n| Connection Monitoring | Specifications for tracking connection status | [Connection Monitoring](specs/connection_monitoring.md) |\n| Circuit Breaking | How the circuit breaker mechanism prevents cascading failures | [Circuit Breaking](specs/circuit_breaking.md) |\n| Self-Healing | Automatic recovery mechanisms for failed connections | [Self-Healing](specs/self_healing.md) |\n| Status Notifications | Event system for connection status changes | [Status Notifications](specs/status_notifications.md) |\n| Failure Isolation | Techniques to isolate failures and protect the application | [Failure Isolation](specs/failure_isolation.md) |\n| Configuration | Configuration options and defaults | [Configuration](specs/configuration.md) |\n| Testing Strategy | Approaches for testing the library | [Testing Strategy](specs/testing_strategy.md) |\n| API | Public interface and usage patterns (includes Decorator API) | [API](specs/api.md) |","ref":"specs.html#specifications-index"},{"type":"extras","title":"Development Roadmap - Brama Library Specifications","doc":"This specification will evolve as development progresses. Implementation priorities:\n\n1. Core connection monitoring and registry functionality\n2. Basic circuit breaking mechanism\n3. Expiry strategies (fixed and progressive)\n4. Status notification system\n5. Function decorator API for seamless integration\n6. Event subscription and delivery system\n7. Telemetry integration\n8. Advanced configuration options\n9. Documentation and examples\n\nThe development will follow an iterative approach, starting with the core functionality and progressively adding features while maintaining backward compatibility.","ref":"specs.html#development-roadmap"},{"type":"extras","title":"Architecture Specification","doc":"# Architecture Specification","ref":"architecture.html"},{"type":"extras","title":"Overview - Architecture Specification","doc":"Brama's architecture is designed to provide robust connection management with minimal overhead. This document outlines the core components and their interactions within the library.","ref":"architecture.html#overview"},{"type":"extras","title":"Core Components - Architecture Specification","doc":"","ref":"architecture.html#core-components"},{"type":"extras","title":"Connection Registry - Architecture Specification","doc":"The central component that maintains the status of all registered connections. It:\n- Stores connection identifiers and their current status\n- Tracks failure counts and circuit states\n- Provides an API for looking up connection statuses","ref":"architecture.html#connection-registry"},{"type":"extras","title":"Circuit Breaker - Architecture Specification","doc":"This component implements the circuit breaking pattern:\n- Monitors connection attempts and failures\n- Implements configurable thresholds for state transitions\n- Prevents calls to failing dependencies","ref":"architecture.html#circuit-breaker"},{"type":"extras","title":"Status Manager - Architecture Specification","doc":"Responsible for:\n- Updating connection status based on feedback\n- Implementing the self-healing mechanism with expiry times\n- Running the periodic cleanup process","ref":"architecture.html#status-manager"},{"type":"extras","title":"Event System - Architecture Specification","doc":"Enables applications to react to connection status changes:\n- Publishes events when connection status changes\n- Allows subscribers to be notified of specific or all connection changes\n- Provides a decoupled way to react to availability changes\n- Maintains a registry of subscribers with filtering preferences\n- Supports subscribing to events by:\n  - Event type (e.g., :state_change, :failure, :success)\n  - Connection identifier\n  - Connection scope\n- Allows unsubscribing from specific or all events\n- Delivers events as standard messages to subscribing processes\n- Implements filter-based delivery to minimize message volume","ref":"architecture.html#event-system"},{"type":"extras","title":"Decorator - Architecture Specification","doc":"Provides a clean way to wrap functions with circuit breaking functionality:\n- Uses the `Decorator` library to implement function decorators with `@decorate` syntax\n- Automatically checks circuit status before function execution\n- Reports success or failure after function execution based on return values\n- Uses customizable error handlers to determine success/failure status:\n  - `:success` - Operation completed successfully\n  - `:failure` - Operation failed\n  - `{:failure, reason}` - Operation failed with specific reason\n  - `:ignore` - Operation result should not affect circuit state\n- Catches and properly handles exceptions by recording failures and re-raising\n- Handles circuit state appropriately on errors\n- Minimizes boilerplate code for API consumers\n- Works with both synchronous and asynchronous functions","ref":"architecture.html#decorator"},{"type":"extras","title":"System Interaction Flow - Architecture Specification","doc":"1. Application registers connections with unique identifiers\n2. Before making an external call, application checks connection status\n3. After a call, application reports success or failure\n4. Circuit breaker updates the status based on success/failure patterns\n5. Subscribers are notified of any status changes","ref":"architecture.html#system-interaction-flow"},{"type":"extras","title":"State Management - Architecture Specification","doc":"Connection states will follow this pattern:\n1. **Closed** - Normal operation, calls allowed\n2. **Open** - Circuit broken, calls rejected without attempting\n3. **Half-Open** - Test state, limited calls allowed to check recovery","ref":"architecture.html#state-management"},{"type":"extras","title":"Circuit Breaker Terminology Note - Architecture Specification","doc":"The circuit breaker pattern borrows terminology from electrical engineering:\n\n- **Closed Circuit**: In electrical terms, a closed circuit allows current to flow through it. Similarly, in our circuit breaker, the \"Closed\" state allows calls to flow through to the external service. This is the normal, healthy operational state.\n\n- **Open Circuit**: An electrical open circuit has a break that prevents current from flowing. In our circuit breaker, the \"Open\" state prevents calls from reaching the failing service, protecting the system from cascading failures.\n\n- **Half-Open Circuit**: This transitional state allows a limited number of test calls to determine if the external service has recovered. Based on the success or failure of these test calls, the circuit will either return to \"Closed\" (if successful) or remain \"Open\" (if still failing).\n\nThis naming convention, while potentially counterintuitive at first, maintains consistency with industry-standard circuit breaker implementations.","ref":"architecture.html#circuit-breaker-terminology-note"},{"type":"extras","title":"Error Handling - Architecture Specification","doc":"- Each component will handle its own internal errors\n- Failures in one connection won't affect others\n- The library will be defensive in handling invalid inputs\n- All public functions will return tagged tuples (`{:ok, result}` or `{:error, reason}`)","ref":"architecture.html#error-handling"},{"type":"extras","title":"Performance Considerations - Architecture Specification","doc":"- Connection status lookups should be constant time operations\n- Status updates should have minimal overhead\n- The library should be able to handle thousands of connections\n- Memory usage should scale linearly with the number of connections","ref":"architecture.html#performance-considerations"},{"type":"extras","title":"Extension Points - Architecture Specification","doc":"Brama's architecture will allow for extension in these areas:\n- Advanced status tracking mechanisms\n- Telemetry integration\n- Custom notification handlers\n- Custom event filtering mechanisms\n- Custom decorator behavior for specialized use cases","ref":"architecture.html#extension-points"},{"type":"extras","title":"Circuit Breaking Specification","doc":"# Circuit Breaking Specification","ref":"circuit_breaking.html"},{"type":"extras","title":"Overview - Circuit Breaking Specification","doc":"Circuit breaking is a key resilience pattern in Brama that prevents cascading failures by stopping requests to failing dependencies. This document outlines how the circuit breaker works, state transitions, and configuration options.","ref":"circuit_breaking.html#overview"},{"type":"extras","title":"Circuit States - Circuit Breaking Specification","doc":"The circuit breaker pattern involves three distinct states:\n\n1. **Closed** (`:closed`)\n   - Normal operation\n   - All requests to the dependency are allowed\n   - Failures are counted\n   - Transitions to open when failure threshold is reached\n\n2. **Open** (`:open`)\n   - Failure state\n   - All requests are immediately rejected without attempting to call the dependency\n   - Failure fast prevents resource exhaustion\n   - Transitions to half-open after expiry time passes\n\n3. **Half-Open** (`:half_open`)\n   - Recovery testing state\n   - A limited number of test requests are allowed\n   - Success transitions to closed\n   - Failure transitions back to open","ref":"circuit_breaking.html#circuit-states"},{"type":"extras","title":"State Transition Rules - Circuit Breaking Specification","doc":"","ref":"circuit_breaking.html#state-transition-rules"},{"type":"extras","title":"Closed to Open - Circuit Breaking Specification","doc":"- Transition occurs when consecutive failures reach `max_attempts` threshold\n- Default `max_attempts` is 10 (configurable)\n- When transitioning to open, a timestamp is recorded for expiry calculation","ref":"circuit_breaking.html#closed-to-open"},{"type":"extras","title":"Open to Half-Open - Circuit Breaking Specification","doc":"- Transition occurs automatically after `expiry` time has passed\n- Default `expiry` is 60,000ms (1 minute, configurable)\n- No requests allowed during open state","ref":"circuit_breaking.html#open-to-half-open"},{"type":"extras","title":"Half-Open to Closed - Circuit Breaking Specification","doc":"- First successful request in half-open state transitions to closed\n- Resets failure counter to 0","ref":"circuit_breaking.html#half-open-to-closed"},{"type":"extras","title":"Half-Open to Open - Circuit Breaking Specification","doc":"- First failed request in half-open state transitions back to open\n- Resets expiry timer","ref":"circuit_breaking.html#half-open-to-open"},{"type":"extras","title":"Configuration - Circuit Breaking Specification","doc":"The circuit breaker can be configured globally:\n\n```elixir\nconfig :brama,\n  max_attempts: 10,         # Attempts before circuit opens\n  expiry: 60_000            # Circuit open duration in ms\n```\n\nOr per-connection during registration:\n\n```elixir\nBrama.register(\"payment_api\", max_attempts: 5, expiry: 30_000)\n```","ref":"circuit_breaking.html#configuration"},{"type":"extras","title":"Implementation Details - Circuit Breaking Specification","doc":"","ref":"circuit_breaking.html#implementation-details"},{"type":"extras","title":"Failure Counting - Circuit Breaking Specification","doc":"- Only consecutive failures count toward threshold\n- A single success resets the consecutive failure counter to 0\n- Different connections track failures separately","ref":"circuit_breaking.html#failure-counting"},{"type":"extras","title":"Fast Circuit Check - Circuit Breaking Specification","doc":"- Circuit status checks are designed to be very fast (constant time)\n- No network or disk I/O during status checks\n- Optimized for high-throughput applications","ref":"circuit_breaking.html#fast-circuit-check"},{"type":"extras","title":"Thread Safety - Circuit Breaking Specification","doc":"- All circuit state operations are atomic\n- Concurrent requests will see consistent circuit state","ref":"circuit_breaking.html#thread-safety"},{"type":"extras","title":"Manual Circuit Control - Circuit Breaking Specification","doc":"In addition to automatic circuit breaking, Brama provides manual control:\n\n```elixir\n# Force open a circuit\nBrama.open_circuit!(identifier, reason: \"Maintenance\", expires_in: 3_600_000)\n\n# Force close a circuit\nBrama.close_circuit!(identifier)\n\n# Reset circuit to initial state\nBrama.reset_circuit!(identifier)\n```","ref":"circuit_breaking.html#manual-circuit-control"},{"type":"extras","title":"Status Notifications Specification","doc":"# Status Notifications Specification","ref":"status_notifications.html"},{"type":"extras","title":"Overview - Status Notifications Specification","doc":"Brama provides a comprehensive event notification system that allows applications to react to connection status changes. This document outlines how the notification system works, the types of events generated, and how to subscribe to and handle these events.","ref":"status_notifications.html#overview"},{"type":"extras","title":"Event Types - Status Notifications Specification","doc":"Brama generates the following types of events:\n\n| Event Type | Description |\n|------------|-------------|\n| `:state_change` | Connection state has changed (e.g., closed to open) |\n| `:registration` | New connection registered with the system |\n| `:failure` | Connection failure reported |\n| `:success` | Connection success reported |\n| `:expiry` | Circuit expiry time reached |\n| `:cleanup` | Connection removed during cleanup |\n| `:circuit_opened` | Circuit has been opened |\n| `:circuit_closed` | Circuit has been closed |\n| `:circuit_reset` | Circuit has been reset |\n| `:connection_registered` | New connection registered |\n| `:connection_unregistered` | Connection unregistered |\n| `:circuit_half_opened` | Circuit has transitioned to half-open |\n| `:connection_removed` | Connection has been removed |","ref":"status_notifications.html#event-types"},{"type":"extras","title":"Event Data Structure - Status Notifications Specification","doc":"Each event contains detailed information about the connection and the event itself:\n\n```elixir\n%{\n  event: :state_change,          # Type of event\n  timestamp: 1630000000000,      # When the event occurred\n  connection: \"payment_api\",     # Connection identifier\n  scope: \"payments\",             # Connection scope\n  data: %{                       # Event-specific data\n    previous_state: :closed,\n    new_state: :open,\n    reason: \"Threshold exceeded\",\n    failure_count: 10\n  }\n}\n```","ref":"status_notifications.html#event-data-structure"},{"type":"extras","title":"Subscription API - Status Notifications Specification","doc":"","ref":"status_notifications.html#subscription-api"},{"type":"extras","title":"Global Subscription - Status Notifications Specification","doc":"Subscribe to all events for all connections:\n\n```elixir\nBrama.subscribe()\n```","ref":"status_notifications.html#global-subscription"},{"type":"extras","title":"Filtered Subscription - Status Notifications Specification","doc":"Subscribe to specific events or connections:\n\n```elixir\n# Subscribe to specific event types\nBrama.subscribe(events: [:state_change, :failure])\n\n# Subscribe to specific connection\nBrama.subscribe(connection: \"payment_api\")\n\n# Subscribe to scope\nBrama.subscribe(scope: \"payments\")\n\n# Combine filters\nBrama.subscribe(\n  events: [:state_change],\n  connection: \"payment_api\"\n)\n```","ref":"status_notifications.html#filtered-subscription"},{"type":"extras","title":"Unsubscribing - Status Notifications Specification","doc":"Unsubscribe using similar patterns:\n\n```elixir\nBrama.unsubscribe()  # Unsubscribe from all\nBrama.unsubscribe(connection: \"payment_api\")  # Specific unsubscribe\n```","ref":"status_notifications.html#unsubscribing"},{"type":"extras","title":"Event Handling - Status Notifications Specification","doc":"Events are delivered as standard Elixir messages to the subscribing process:\n\n```elixir\ndef handle_info({:brama_event, event}, state) do\n  # Handle the event\n  Logger.info(\"Connection #{event.connection} changed to #{event.data.new_state}\")\n  {:noreply, state}\nend\n```","ref":"status_notifications.html#event-handling"},{"type":"extras","title":"Implementation Details - Status Notifications Specification","doc":"","ref":"status_notifications.html#implementation-details"},{"type":"extras","title":"Event Manager - Status Notifications Specification","doc":"- Implemented using Elixir's Registry or GenStage (depending on volume needs)\n- Maintains a list of subscribers and their filter preferences\n- Efficiently distributes events only to interested subscribers","ref":"status_notifications.html#event-manager"},{"type":"extras","title":"Performance Considerations - Status Notifications Specification","doc":"- Event generation is asynchronous to prevent blocking\n- Subscription filters are applied at the source to minimize message volume\n- High-volume systems can use batched event delivery","ref":"status_notifications.html#performance-considerations"},{"type":"extras","title":"Notification Behavior - Status Notifications Specification","doc":"Events have the following delivery guarantees:\n\n- At-least-once delivery for critical state changes\n- No guaranteed order of events\n- No persistence of events (transient in-memory only)","ref":"status_notifications.html#notification-behavior"},{"type":"extras","title":"Integration with Monitoring Systems - Status Notifications Specification","doc":"For integration with external monitoring systems, Brama provides:\n\n```elixir\n# Custom event handler for integration\nBrama.subscribe(handler: MyApp.MonitoringHandler)\n\n# Telemetry integration\n:telemetry.execute(\n  [:brama, :connection, :state_change],\n  %{duration: System.monotonic_time() - start_time},\n  %{connection: \"payment_api\", previous_state: :closed, new_state: :open}\n)\n```","ref":"status_notifications.html#integration-with-monitoring-systems"},{"type":"extras","title":"Example Usage - Status Notifications Specification","doc":"```elixir\ndefmodule MyApp.ConnectionMonitor do\n  use GenServer\n\n  def start_link(_) do\n    GenServer.start_link(__MODULE__, %{})\n  end\n\n  def init(state) do\n    # Subscribe to all state changes\n    Brama.subscribe(events: [:state_change])\n    {:ok, state}\n  end\n\n  def handle_info({:brama_event, event}, state) do\n    if event.data.new_state == :open do\n      # Alert on circuit open\n      MyApp.Alerts.send(\"Circuit opened for #{event.connection}\")\n    end\n    \n    {:noreply, state}\n  end\nend\n```","ref":"status_notifications.html#example-usage"},{"type":"extras","title":"Connection Monitoring Specification","doc":"# Connection Monitoring Specification","ref":"connection_monitoring.html"},{"type":"extras","title":"Overview - Connection Monitoring Specification","doc":"Connection monitoring is a core feature of Brama that tracks the status of connections to external dependencies. This document outlines how connection monitoring works, its data structures, and key functions.","ref":"connection_monitoring.html#overview"},{"type":"extras","title":"Connection Identification - Connection Monitoring Specification","doc":"Each connection is uniquely identified by:\n- A string identifier (required) - Typically representing the service name\n- A scope (optional) - For grouping related connections (e.g., \"payments\", \"inventory\")\n\nThis combination allows for precise tracking and management of multiple connections to the same service.","ref":"connection_monitoring.html#connection-identification"},{"type":"extras","title":"Connection States - Connection Monitoring Specification","doc":"Connections can be in one of the following states:\n\n| State | Description |\n|-------|-------------|\n| `:closed` | Default state, connection is operational |\n| `:open` | Circuit is open, connection is not accepting requests |\n| `:half_open` | Transitional state to test if the connection has recovered |\n| `:unknown` | Initial state before any data has been collected |","ref":"connection_monitoring.html#connection-states"},{"type":"extras","title":"Data Structure - Connection Monitoring Specification","doc":"Each connection will be stored with the following information:\n\n```elixir\n%{\n  identifier: \"service_name\",\n  scope: \"system\",             # Optional, default is nil\n  state: :closed,              # Current circuit state\n  failure_count: 0,            # Number of consecutive failures\n  last_failure_time: nil,      # Timestamp of last failure\n  last_success_time: nil,      # Timestamp of last success\n  opened_at: nil,              # When the circuit was last opened\n  metadata: %{}                # Optional user-defined metadata\n}\n```","ref":"connection_monitoring.html#data-structure"},{"type":"extras","title":"Core Functions - Connection Monitoring Specification","doc":"","ref":"connection_monitoring.html#core-functions"},{"type":"extras","title":"Registration - Connection Monitoring Specification","doc":"```elixir\nBrama.register(identifier, opts \\\\ [])\n```\n\nRegisters a new connection with the system. Options include:\n- `scope`: Grouping category (default: `nil`)\n- `metadata`: Custom information about the connection","ref":"connection_monitoring.html#registration"},{"type":"extras","title":"Status Checking - Connection Monitoring Specification","doc":"```elixir\nBrama.available?(identifier, opts \\\\ [])\n```\n\nChecks if a connection is available (circuit closed). Returns:\n- `true` - Connection is available\n- `false` - Connection is unavailable (circuit open)","ref":"connection_monitoring.html#status-checking"},{"type":"extras","title":"Status Reporting - Connection Monitoring Specification","doc":"```elixir\nBrama.success(identifier, opts \\\\ [])\nBrama.failure(identifier, opts \\\\ [])\n```\n\nThese functions report connection success or failure, updating internal counters and potentially changing the circuit state.","ref":"connection_monitoring.html#status-reporting"},{"type":"extras","title":"Status Retrieval - Connection Monitoring Specification","doc":"```elixir\nBrama.status(identifier, opts \\\\ [])\n```\n\nReturns detailed status information about a connection:\n```elixir\n{:ok, %{state: :closed, failure_count: 0, ...}}\n```","ref":"connection_monitoring.html#status-retrieval"},{"type":"extras","title":"Connection Lifecycle - Connection Monitoring Specification","doc":"1. Connection is registered with `Brama.register/2`\n2. Application checks availability with `Brama.available?/2` before making external calls\n3. Application reports success or failure with `Brama.success/2` or `Brama.failure/2`\n4. If failure threshold is reached, circuit opens\n5. After expiry time, circuit transitions to half-open\n6. First successful call in half-open state closes the circuit","ref":"connection_monitoring.html#connection-lifecycle"},{"type":"extras","title":"Monitoring Metrics - Connection Monitoring Specification","doc":"The following metrics will be available for each connection:\n- Current state\n- Total failure count\n- Consecutive failure count\n- Time since last state change\n- Success rate (ratio of successful to total calls)","ref":"connection_monitoring.html#monitoring-metrics"},{"type":"extras","title":"Connection Cleanup - Connection Monitoring Specification","doc":"Inactive connections can be removed using:\n```elixir\nBrama.unregister(identifier, opts \\\\ [])\n```\n\nAdditionally, the system will periodically clean up connections that have been inactive for a configurable period.","ref":"connection_monitoring.html#connection-cleanup"},{"type":"extras","title":"Self-Healing Specification","doc":"# Self-Healing Specification","ref":"self_healing.html"},{"type":"extras","title":"Overview - Self-Healing Specification","doc":"Self-healing is a critical feature of Brama that allows the system to automatically recover from failures without manual intervention. This document outlines how the self-healing mechanisms work, their configuration, and implementation details.","ref":"self_healing.html#overview"},{"type":"extras","title":"Self-Healing Mechanisms - Self-Healing Specification","doc":"Brama implements several self-healing mechanisms:\n\n1. **Expiry-Based Reset**\n   - Open circuits automatically transition to half-open after a configurable expiry time\n   - This allows the system to attempt recovery without manual intervention\n\n2. **Automatic State Cleanup**\n   - Inactive connections are periodically checked and potentially cleaned up\n   - Prevents accumulation of stale connection data\n\n3. **Progressive Recovery**\n   - Half-open state allows a controlled test of dependency health\n   - Prevents rapid cycle between open and closed states","ref":"self_healing.html#self-healing-mechanisms"},{"type":"extras","title":"Expiry Configuration - Self-Healing Specification","doc":"The expiry time can be configured globally:\n\n```elixir\nconfig :brama,\n  expiry: 60_000  # Default: 1 minute in milliseconds\n```\n\nOr per-connection:\n\n```elixir\nBrama.register(\"payment_api\", expiry: 30_000)  # 30 seconds\n```","ref":"self_healing.html#expiry-configuration"},{"type":"extras","title":"Cleanup Configuration - Self-Healing Specification","doc":"The cleanup process can be configured:\n\n```elixir\nconfig :brama,\n  cleanup_interval: 10_000,     # How often to run cleanup (ms)\n  inactive_threshold: 86_400_000  # Consider inactive after 24 hours\n```","ref":"self_healing.html#cleanup-configuration"},{"type":"extras","title":"Recovery Process - Self-Healing Specification","doc":"","ref":"self_healing.html#recovery-process"},{"type":"extras","title":"Automatic Circuit Reset - Self-Healing Specification","doc":"1. Circuit opens due to failures reaching threshold\n2. Time passes (expiry duration)\n3. Circuit automatically transitions to half-open\n4. Next request becomes a test request\n5. If test succeeds, circuit closes\n6. If test fails, circuit reopens with reset expiry timer","ref":"self_healing.html#automatic-circuit-reset"},{"type":"extras","title":"Connection Cleanup - Self-Healing Specification","doc":"1. Periodic process runs at `cleanup_interval`\n2. Identifies connections with no activity for `inactive_threshold`\n3. Marks these connections for potential cleanup\n4. After extended inactivity, removes connection data","ref":"self_healing.html#connection-cleanup"},{"type":"extras","title":"Implementation Details - Self-Healing Specification","doc":"","ref":"self_healing.html#implementation-details"},{"type":"extras","title":"Time Tracking - Self-Healing Specification","doc":"- Each connection stores timestamps for key events:\n  - Last failure time\n  - Last success time\n  - Open circuit time\n  - Last activity time","ref":"self_healing.html#time-tracking"},{"type":"extras","title":"Expiry Calculation - Self-Healing Specification","doc":"- Current time is compared with `opened_at` timestamp\n- If difference exceeds `expiry`, circuit transitions to half-open","ref":"self_healing.html#expiry-calculation"},{"type":"extras","title":"Cleanup Process - Self-Healing Specification","doc":"- Implemented as a separate process that wakes at configured intervals\n- Uses a low-priority background job to prevent impact on performance\n- Only reviews connections that have exceeded the inactivity threshold","ref":"self_healing.html#cleanup-process"},{"type":"extras","title":"Manual Override - Self-Healing Specification","doc":"While self-healing is automatic, manual controls are available:\n\n```elixir\n# Force immediate reset to half-open\nBrama.reset_expiry!(identifier)\n\n# Set custom expiry for specific connection\nBrama.set_expiry!(identifier, 120_000)  # 2 minutes\n\n# Disable expiry (circuit stays open until manual intervention)\nBrama.disable_expiry!(identifier)\n```","ref":"self_healing.html#manual-override"},{"type":"extras","title":"Progressive Backoff - Self-Healing Specification","doc":"For advanced self-healing, a progressive backoff mechanism is available:\n\n```elixir\nBrama.register(\"flaky_service\", \n  expiry_strategy: :progressive,\n  initial_expiry: 10_000,\n  max_expiry: 300_000,\n  backoff_factor: 2.0\n)\n```\n\nThis strategy:\n- Starts with a short expiry time for first failure\n- Increases expiry geometrically with repeated failures\n- Caps at a maximum expiry time\n- Resets to initial expiry after successful recovery","ref":"self_healing.html#progressive-backoff"},{"type":"extras","title":"Failure Isolation Specification","doc":"# Failure Isolation Specification","ref":"failure_isolation.html"},{"type":"extras","title":"Overview - Failure Isolation Specification","doc":"Failure isolation is a key resilience pattern in Brama that prevents failures in one external dependency from affecting the entire system. This document outlines how Brama implements failure isolation, its benefits, and usage patterns.","ref":"failure_isolation.html#overview"},{"type":"extras","title":"Failure Isolation Principles - Failure Isolation Specification","doc":"Brama follows these core principles for failure isolation:\n\n1. **Independence** - Each connection is tracked independently\n2. **Containment** - Failures in one system don't impact others\n3. **Fast Rejection** - Failing connections are rejected quickly without consuming resources\n4. **Degraded Functionality** - Applications can continue with reduced functionality","ref":"failure_isolation.html#failure-isolation-principles"},{"type":"extras","title":"Isolation Mechanisms - Failure Isolation Specification","doc":"","ref":"failure_isolation.html#isolation-mechanisms"},{"type":"extras","title":"Connection Isolation - Failure Isolation Specification","doc":"- Each connection is tracked separately\n- Failure of one connection doesn't affect others\n- Different services can have different circuit thresholds and expiry times","ref":"failure_isolation.html#connection-isolation"},{"type":"extras","title":"Scope-Based Isolation - Failure Isolation Specification","doc":"- Connections can be grouped by logical scope (e.g., \"payments\", \"inventory\")\n- Enables business-domain isolation\n- Status querying and reporting by scope","ref":"failure_isolation.html#scope-based-isolation"},{"type":"extras","title":"Implementation - Failure Isolation Specification","doc":"","ref":"failure_isolation.html#implementation"},{"type":"extras","title":"Independent Circuit State - Failure Isolation Specification","doc":"Each connection maintains its own circuit state:\n\n```elixir\n# Different services have independent circuits\nBrama.register(\"payment_api\")\nBrama.register(\"shipping_api\")\n\n# Failure in payment_api doesn't affect shipping_api\nBrama.failure(\"payment_api\")\nBrama.available?(\"shipping_api\")  # Still true\n```","ref":"failure_isolation.html#independent-circuit-state"},{"type":"extras","title":"Failure Handling - Failure Isolation Specification","doc":"Applications use Brama for graceful degradation:\n\n```elixir\ndefmodule UserService do\n  use Brama.Decorator\n\n  def get_user_data(user_id) do\n    profile = get_user_profile(user_id)\n    case get_payment_history(user_id) do\n      {:ok, history} ->\n        # ...\n      {:error, _reason} ->\n        # ...\n    end\n  end\n  \n  # Circuit breaker automatically handles checking availability,\n  # reporting success/failure, and providing fallback\n  @circuit_breaker identifier: \"payment_api\"\n  defp get_payment_history(user_id) do\n    # Fetch data from an external API\n  end\n  \n  defp get_user_profile(user_id) do\n    # Profile fetching logic\n  end\nend\n```","ref":"failure_isolation.html#failure-handling"},{"type":"extras","title":"Group Management - Failure Isolation Specification","doc":"For logical service groups:\n\n```elixir\n# Register a service group\nBrama.register_group(\"payment_services\", [\n  \"payment_api\", \n  \"invoice_service\", \n  \"tax_calculator\"\n])\n\n# Check entire group\nBrama.group_available?(\"payment_services\")\n\n# Report to entire group\nBrama.group_failure(\"payment_services\")\n```","ref":"failure_isolation.html#group-management"},{"type":"extras","title":"Isolation Benefits - Failure Isolation Specification","doc":"The failure isolation provided by Brama offers these benefits:\n\n1. **Improved Stability** - Prevents cascading failures\n2. **Resource Conservation** - Fast-fails requests to unavailable services\n3. **Partial Availability** - System continues operating with reduced functionality\n4. **Targeted Recovery** - Services can recover independently","ref":"failure_isolation.html#isolation-benefits"},{"type":"extras","title":"Failure Isolation Strategy Recommendations - Failure Isolation Specification","doc":"1. Identify critical vs. non-critical dependencies\n2. Use shorter thresholds for critical services\n3. Always provide degraded functionality where possible\n4. Use scopes to group related services\n5. Consider fallbacks for critical functionality","ref":"failure_isolation.html#failure-isolation-strategy-recommendations"},{"type":"extras","title":"Configuration Specification","doc":"# Configuration Specification","ref":"configuration.html"},{"type":"extras","title":"Overview - Configuration Specification","doc":"Brama's configuration system is designed to be flexible while providing reasonable defaults. This document outlines the configuration options, their defaults, and how to customize them for specific needs.","ref":"configuration.html#overview"},{"type":"extras","title":"Global Configuration - Configuration Specification","doc":"Brama is configured in your application's config files:\n\n```elixir\n# In config/config.exs\nconfig :brama,\n  # Circuit breaker settings\n  max_attempts: 5,            # Attempts before circuit opens (default: 5)\n  expiry: 60_000,             # Circuit open duration in ms (default: 60,000)\n  \n  # Cleanup settings\n  cleanup_interval: 10_000,   # Status check interval in ms (default: 10,000)\n  inactive_threshold: 86_400_000,  # Time before cleanup in ms (default: 24 hours)\n```","ref":"configuration.html#global-configuration"},{"type":"extras","title":"Per-Connection Configuration - Configuration Specification","doc":"Each connection can override global settings during registration:\n\n```elixir\nBrama.register(\"payment_api\", \n  max_attempts: 5,            # Override max attempts for this connection\n  expiry: 30_000,             # Override expiry time for this connection\n  scope: \"payment_services\",  # Logical grouping\n  metadata: %{                # Additional information\n    url: \"https://payments.example.com/api\",\n    priority: :critical\n  }\n)\n```","ref":"configuration.html#per-connection-configuration"},{"type":"extras","title":"Environment-Specific Configurations - Configuration Specification","doc":"Brama supports different configurations per environment:\n\n```elixir\n# In config/dev.exs\nconfig :brama,\n  max_attempts: 3,            # Lower threshold in development\n  expiry: 10_000              # Faster recovery in development\n\n# In config/test.exs\nconfig :brama,\n  testing_mode: true,         # Enable testing mode\n  max_attempts: 1,            # Immediate circuit opening for tests\n  cleanup_interval: 1_000     # Faster cleanup for tests\n\n# In config/prod.exs\nconfig :brama,\n  max_attempts: 10,           # Standard threshold in production\n  expiry: 60_000,             # Standard recovery in production\n  cleanup_interval: 30_000    # Less frequent cleanup in production\n```","ref":"configuration.html#environment-specific-configurations"},{"type":"extras","title":"Configuration Precedence - Configuration Specification","doc":"Configuration values are determined in this order of precedence:\n1. Per-connection settings (highest priority)\n2. Runtime configuration updates\n3. Environment-specific global settings\n4. Default values (lowest priority)","ref":"configuration.html#configuration-precedence"},{"type":"extras","title":"Runtime Configuration - Configuration Specification","doc":"Brama allows updating configuration at runtime:\n\n```elixir\n# Update global settings\nBrama.configure(max_attempts: 15, expiry: 120_000)\n\n# Update specific connection\nBrama.configure(\"payment_api\", max_attempts: 5, expiry: 30_000)\n```","ref":"configuration.html#runtime-configuration"},{"type":"extras","title":"Configuration Options Reference - Configuration Specification","doc":"| Option | Type | Default | Description |\n|--------|------|---------|-------------|\n| `max_attempts` | Integer | 5 | Number of failures before circuit opens |\n| `expiry` | Integer | 60_000 | Time in ms before open circuit transitions to half-open |\n| `cleanup_interval` | Integer | 10_000 | Time in ms between cleanup checks |\n| `inactive_threshold` | Integer | 86_400_000 | Time in ms before connection considered inactive |\n| `telemetry_prefix` | List | `[:brama]` | Prefix for telemetry events |\n| `testing_mode` | Boolean | `false` | Enable testing features |","ref":"configuration.html#configuration-options-reference"},{"type":"extras","title":"Advanced Configuration - Configuration Specification","doc":"","ref":"configuration.html#advanced-configuration"},{"type":"extras","title":"Progressive Backoff - Configuration Specification","doc":"```elixir\nBrama.register(\"flaky_service\", \n  expiry_strategy: :progressive,\n  initial_expiry: 10_000,\n  max_expiry: 300_000,\n  backoff_factor: 2.0\n)\n```","ref":"configuration.html#progressive-backoff"},{"type":"extras","title":"Configuration at Application Start - Configuration Specification","doc":"Initialize Brama with dynamic configuration at application start:\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    # Initial configuration based on environment\n    Brama.configure(\n      max_attempts: Application.get_env(:my_app, :circuit_max_attempts, 10),\n      expiry: Application.get_env(:my_app, :circuit_expiry, 60_000)\n    )\n    \n    # ... supervisor setup\n  end\nend\n```","ref":"configuration.html#configuration-at-application-start"},{"type":"extras","title":"Configuration Validation - Configuration Specification","doc":"All configuration values are validated at startup and when changed:\n- Numeric values must be positive integers\n- Timeout values have reasonable minimums\n- Invalid configurations raise helpful error messages","ref":"configuration.html#configuration-validation"},{"type":"extras","title":"Testing Strategy Specification","doc":"# Testing Strategy Specification","ref":"testing_strategy.html"},{"type":"extras","title":"Overview - Testing Strategy Specification","doc":"This document outlines the testing strategy for the Brama library, focusing on unit testing, integration testing, and providing helper functions for testing applications that use Brama.","ref":"testing_strategy.html#overview"},{"type":"extras","title":"Testing Levels - Testing Strategy Specification","doc":"","ref":"testing_strategy.html#testing-levels"},{"type":"extras","title":"Unit Testing - Testing Strategy Specification","doc":"All components of Brama will have comprehensive unit tests:\n\nEach unit test will focus on a single component in isolation, using mocks or stubs for dependencies.","ref":"testing_strategy.html#unit-testing"},{"type":"extras","title":"Integration Testing - Testing Strategy Specification","doc":"Integration tests will verify interactions between components:\n\n- Connection registration to event notification flow\n- Circuit state transitions under various conditions\n- Cleanup and expiry interactions","ref":"testing_strategy.html#integration-testing"},{"type":"extras","title":"Test Helpers Module - Testing Strategy Specification","doc":"Brama provides a dedicated `Brama.TestHelpers` module that simplifies testing of applications that use Brama. This module includes functions for:\n\n- Directly manipulating circuit states\n- Simulating failures and recoveries\n- Working with connection scopes","ref":"testing_strategy.html#test-helpers-module"},{"type":"extras","title":"Available Helper Functions - Testing Strategy Specification","doc":"#### Setting Circuit State\n\n```elixir\n@spec set_state(String.t(), atom(), Keyword.t()) :: :ok | {:error, term()}\n```\n\nDirectly sets the state of a circuit to `:closed`, `:open`, or `:half_open`. Options include:\n\n- `scope`: Optional scope to target a specific connection scope\n- `reason`: Optional reason for the state change (default: \"Test setup\")\n\nThe implementation uses the appropriate Brama functions internally:\n- For `:open` state, uses `Brama.open_circuit!`\n- For `:closed` state, uses `Brama.close_circuit!`\n- For `:half_open` state, opens the circuit first and then uses a specialized internal function\n\n```elixir\n# Force a connection into open state\nBrama.TestHelpers.set_state(\"payment_api\", :open)\n\n# Force a connection in a specific scope into closed state\nBrama.TestHelpers.set_state(\"payment_api\", :closed, scope: \"external_vendors\")\n\n# Set a connection to half-open state\nBrama.TestHelpers.set_state(\"payment_api\", :half_open)\n```\n\n#### Simulating Failures\n\n```elixir\n@spec add_failures(String.t(), non_neg_integer(), Keyword.t()) :: :ok | {:error, term()}\n```\n\nAdds a specific number of consecutive failures to a connection. Options include:\n\n- `scope`: Optional scope to target a specific connection scope\n- Other options passed to underlying Brama functions\n\n```elixir\n# Add 5 failures to a connection\nBrama.TestHelpers.add_failures(\"payment_api\", 5)\n\n# Add failures to a scoped connection\nBrama.TestHelpers.add_failures(\"payment_api\", 3, scope: \"external_vendors\")\n```\n\n```elixir\n@spec simulate_failures(String.t(), non_neg_integer(), Keyword.t()) :: :ok | {:error, term()}\n```\n\nAlias for `add_failures/3` that simulates a service temporarily failing.\n\n```elixir\n# Simulate 5 failures for a service\nBrama.TestHelpers.simulate_failures(\"payment_api\", 5)\n```\n\n#### Simulating Recovery\n\n```elixir\n@spec simulate_recovery(String.t(), Keyword.t()) :: :ok | {:error, term()}\n```\n\nSimulates a service recovering by reporting a success. Options include:\n\n- `scope`: Optional scope to target a specific connection scope\n- Other options passed to underlying Brama functions\n\n```elixir\n# Simulate service recovery\nBrama.TestHelpers.simulate_recovery(\"payment_api\")\n\n# Simulate recovery for a scoped connection\nBrama.TestHelpers.simulate_recovery(\"payment_api\", scope: \"external_vendors\")\n```","ref":"testing_strategy.html#available-helper-functions"},{"type":"extras","title":"Testing Strategies for Applications - Testing Strategy Specification","doc":"","ref":"testing_strategy.html#testing-strategies-for-applications"},{"type":"extras","title":"Unit Testing with Brama - Testing Strategy Specification","doc":"For applications using Brama, use these testing strategies:\n\n```elixir\n# Example test with Brama\ndefmodule MyApp.PaymentServiceTest do\n  use ExUnit.Case\n  import Brama.TestHelpers\n\n  setup do\n    # Register test connection\n    Brama.register(\"payment_api\")\n    :ok\n  end\n\n  test \"handles payment service unavailability gracefully\" do\n    # Force circuit open\n    set_state(\"payment_api\", :open)\n    \n    # Call code that should handle unavailability\n    result = MyApp.PaymentService.process_payment(100)\n    \n    # Assert correct fallback behavior\n    assert result == {:error, :service_unavailable}\n  end\n\n  test \"circuit opens after multiple failures\" do\n    # Simulate failures\n    add_failures(\"payment_api\", 10)\n    \n    # Assert circuit state\n    assert {:ok, %{state: :open}} = Brama.status(\"payment_api\")\n  end\n\n  test \"circuit transitions to half-open after recovery period\" do\n    # Setup: open the circuit\n    set_state(\"payment_api\", :open)\n    \n    # Simulate time passing (in a real test, you might use mocking for time)\n    # Then transition to half-open\n    set_state(\"payment_api\", :half_open)\n    \n    # Assert correct state\n    assert {:ok, %{state: :half_open}} = Brama.status(\"payment_api\")\n  end\n\n  test \"circuit closes after successful recovery\" do\n    # Setup: set to half-open\n    set_state(\"payment_api\", :half_open)\n    \n    # Simulate recovery\n    simulate_recovery(\"payment_api\")\n    \n    # Assert circuit closed\n    assert {:ok, %{state: :closed}} = Brama.status(\"payment_api\")\n  end\nend\n```","ref":"testing_strategy.html#unit-testing-with-brama"},{"type":"extras","title":"API Specification","doc":"# API Specification","ref":"api.html"},{"type":"extras","title":"Overview - API Specification","doc":"This document specifies the public API for the Brama library. It details the functions available to client applications, their parameters, return values, and usage examples.","ref":"api.html#overview"},{"type":"extras","title":"Core API Functions - API Specification","doc":"","ref":"api.html#core-api-functions"},{"type":"extras","title":"Connection Management - API Specification","doc":"#### Registration\n\n```elixir\n@spec register(identifier :: String.t(), opts :: Keyword.t()) :: {:ok, term()} | {:error, term()}\ndef register(identifier, opts \\\\ [])\n```\n\nRegisters a new connection with Brama.\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n  - `max_attempts` - Override default threshold (default: from config)\n  - `expiry` - Override default expiry time (default: from config)\n  - `expiry_strategy` - Strategy for circuit expiry (`:fixed` or `:progressive`)\n  - `initial_expiry` - Initial expiry time for progressive strategy (in ms)\n  - `max_expiry` - Maximum expiry time for progressive strategy (in ms)\n  - `backoff_factor` - Multiplier for progressive backoff (e.g., 2.0)\n  - `metadata` - Additional information (default: `%{}`)\n\n**Expiry Strategies:**\n- `:fixed` - Always uses the same expiry time (default)\n- `:progressive` - Increases expiry time after each failure using formula: `min(max_expiry, initial_expiry * (backoff_factor ^ failure_count))`\n\n**Returns:**\n- `{:ok, connection_data}` - Successfully registered\n- `{:error, reason}` - Registration failed\n\n**Example:**\n```elixir\n# Basic registration\nBrama.register(\"payment_api\", \n  type: :http, \n  scope: \"payment_services\",\n  max_attempts: 5,\n  expiry: 30_000\n)\n\n# Registration with progressive backoff\nBrama.register(\"flaky_service\",\n  scope: \"external_apis\",\n  expiry_strategy: :progressive,\n  initial_expiry: 5_000,\n  max_expiry: 120_000,\n  backoff_factor: 2.0\n)\n```\n\n#### Unregistration\n\n```elixir\n@spec unregister(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef unregister(identifier, opts \\\\ [])\n```\n\nRemoves a connection from Brama.\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n\n**Returns:**\n- `:ok` - Successfully unregistered\n- `{:error, :not_found}` - Connection not found\n\n**Example:**\n```elixir\nBrama.unregister(\"payment_api\")\n```","ref":"api.html#connection-management"},{"type":"extras","title":"Status Checking - API Specification","doc":"```elixir\n@spec available?(identifier :: String.t(), opts :: Keyword.t()) :: boolean()\ndef available?(identifier, opts \\\\ [])\n```\n\nChecks if a connection is available (circuit closed).\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n\n**Returns:**\n- `true` - Connection is available\n- `false` - Connection is unavailable or not found\n\n**Example:**\n```elixir\nif Brama.available?(\"payment_api\") do\n  # Make API call\nelse\n  # Use fallback\nend\n```","ref":"api.html#status-checking"},{"type":"extras","title":"Status Reporting - API Specification","doc":"```elixir\n@spec success(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef success(identifier, opts \\\\ [])\n\n@spec failure(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef failure(identifier, opts \\\\ [])\n```\n\nReport connection success or failure.\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n  - `reason` - Optional reason for failure (for `failure/2` only)\n\n**Returns:**\n- `:ok` - Status updated\n- `{:error, :not_found}` - Connection not found\n\n**Example:**\n```elixir\ncase make_api_call() do\n  {:ok, result} -> \n    Brama.success(\"payment_api\")\n    {:ok, result}\n  {:error, reason} -> \n    Brama.failure(\"payment_api\", reason: inspect(reason))\n    {:error, reason}\nend\n```","ref":"api.html#status-reporting"},{"type":"extras","title":"Status Retrieval - API Specification","doc":"```elixir\n@spec status(identifier :: String.t(), opts :: Keyword.t()) :: {:ok, map()} | {:error, term()}\ndef status(identifier, opts \\\\ [])\n```\n\nGet detailed status for a connection.\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n\n**Returns:**\n- `{:ok, status_map}` - Status information\n- `{:error, :not_found}` - Connection not found\n\n**Example:**\n```elixir\n{:ok, status} = Brama.status(\"payment_api\")\nIO.puts(\"Connection state: #{status.state}, Failures: #{status.failure_count}\")\n```","ref":"api.html#status-retrieval"},{"type":"extras","title":"Circuit Control API - API Specification","doc":"","ref":"api.html#circuit-control-api"},{"type":"extras","title":"Manual Circuit Control - API Specification","doc":"```elixir\n@spec open_circuit!(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef open_circuit!(identifier, opts \\\\ [])\n\n@spec close_circuit!(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef close_circuit!(identifier, opts \\\\ [])\n\n@spec reset_circuit!(identifier :: String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef reset_circuit!(identifier, opts \\\\ [])\n```\n\nManually control circuit state.\n\n**Parameters:**\n- `identifier` - String identifier for the connection\n- `opts` - Options including:\n  - `type` - Connection type (default: `:general`)\n  - `scope` - Grouping category (default: `nil`)\n  - `reason` - Optional reason for state change\n  - `expires_in` - Optional custom expiry (for `open_circuit!/2` only)\n\n**Returns:**\n- `:ok` - Circuit state updated\n- `{:error, reason}` - Operation failed\n\n**Example:**\n```elixir\n# Force open for maintenance\nBrama.open_circuit!(\"payment_api\", \n  reason: \"Scheduled maintenance\", \n  expires_in: 3_600_000\n)\n\n# Force close after maintenance\nBrama.close_circuit!(\"payment_api\")\n```","ref":"api.html#manual-circuit-control"},{"type":"extras","title":"Event System API - API Specification","doc":"","ref":"api.html#event-system-api"},{"type":"extras","title":"Subscription Management - API Specification","doc":"```elixir\n@spec subscribe(opts :: Keyword.t()) :: {:ok, reference()} | {:error, term()}\ndef subscribe(opts \\\\ [])\n\n@spec unsubscribe(opts :: Keyword.t()) :: :ok | {:error, term()}\ndef unsubscribe(opts \\\\ [])\n```\n\nSubscribe/unsubscribe to connection events.\n\n**Parameters:**\n- `opts` - Filter options including:\n  - `events` - List of event types (default: all)\n  - `connection` - Connection identifier (default: all)\n  - `scope` - Connection scope (default: all)\n  - `handler` - Custom handler module (default: send messages to current process)\n\n**Returns:**\n- `{:ok, subscription_reference}` - Subscription successful\n- `{:error, reason}` - Subscription failed\n\n**Example:**\n```elixir\n# Subscribe to all state changes\n{:ok, _ref} = Brama.subscribe(events: [:state_change])\n\n# Handle events\ndef handle_info({:brama_event, event}, state) do\n  Logger.info(\"Connection #{event.connection} changed to #{event.data.new_state}\")\n  {:noreply, state}\nend\n```","ref":"api.html#subscription-management"},{"type":"extras","title":"Configuration API - API Specification","doc":"```elixir\n@spec configure(config :: Keyword.t() | String.t(), opts :: Keyword.t()) :: :ok | {:error, term()}\ndef configure(config, opts \\\\ [])\n```\n\nUpdate configuration at runtime.\n\n**Parameters:**\n- `config` - Configuration keywords or connection identifier\n- `opts` - Configuration options when first parameter is a connection\n\n**Returns:**\n- `:ok` - Configuration updated\n- `{:error, reason}` - Update failed\n\n**Example:**\n```elixir\n# Update global settings\nBrama.configure(max_attempts: 10, expiry: 60_000)\n\n# Update specific connection\nBrama.configure(\"payment_api\",\n  max_attempts: 5,\n  expiry: 30_000,\n  expiry_strategy: :progressive,\n  initial_expiry: 5_000,\n  max_expiry: 60_000,\n  backoff_factor: 2.0\n)\n```","ref":"api.html#configuration-api"},{"type":"extras","title":"Expiry Strategies - API Specification","doc":"Brama supports different strategies for determining how long a circuit remains open:\n\n#### Fixed Expiry (Default)\n\nThe circuit remains open for a fixed amount of time before transitioning to half-open:\n\n```elixir\nBrama.configure(\"payment_api\",\n  expiry_strategy: :fixed,  # This is the default strategy\n  expiry: 30_000            # 30 seconds\n)\n```\n\n#### Progressive Backoff\n\nThe circuit's expiry time increases with each consecutive failure, allowing for more sophisticated self-healing:\n\n```elixir\nBrama.configure(\"flaky_service\",\n  expiry_strategy: :progressive,\n  initial_expiry: 5_000,     # Start with 5 seconds\n  max_expiry: 300_000,       # Cap at 5 minutes\n  backoff_factor: 2.0        # Double the time with each failure\n)\n```\n\nWith this configuration, expiry times would progress: 5s → 10s → 20s → 40s → 80s → 160s → 300s (capped at max)","ref":"api.html#expiry-strategies"},{"type":"extras","title":"Decorator API - API Specification","doc":"The decorator API uses the [Decorator](https://hexdocs.pm/decorator/) library to provide a clean way to wrap functions with circuit breaking logic.","ref":"api.html#decorator-api"},{"type":"extras","title":"Usage - API Specification","doc":"```elixir\ndefmodule PaymentService do\n  use Brama.Decorator\n\n  @decorate circuit_breaker(identifier: \"payment_api\")\n  def process_payment(payment) do\n    PaymentAPI.process(payment)\n  end\nend\n```","ref":"api.html#usage"},{"type":"extras","title":"Decorator Options - API Specification","doc":"The circuit breaker decorator accepts the following options:\n\n- `identifier` - A unique string identifier for the circuit breaker instance (required)\n- `error_handler` - Custom function to handle responses and determine success/failure status (optional)","ref":"api.html#decorator-options"},{"type":"extras","title":"Error Handling - API Specification","doc":"The error handler function can return the following statuses:\n\n- `:success` - Operation completed successfully, keeps circuit closed\n- `:failure` - Operation failed, increments failure count\n- `{:failure, reason}` - Operation failed with specific reason\n- `:ignore` - Operation result should not affect circuit breaker state\n\nDefault error handling behavior:\n\n```elixir\n@decorate circuit_breaker(\n  identifier: \"payment_api\",\n  error_handler: fn\n    {:ok, _} -> :success\n    {:error, reason} -> {:failure, reason}\n    _ -> :ignore\n  end\n)\ndef process_payment(payment) do\n  PaymentAPI.process(payment)\nend\n```","ref":"api.html#error-handling"},{"type":"extras","title":"Service Unavailability - API Specification","doc":"When the circuit is open:\n1. The decorated function is not executed\n2. Returns `{:error, status}` where status contains circuit breaker state\n3. Prevents overwhelming failing services with requests\n\nExample response when circuit is open:\n```elixir\n{:error, status} # where status contains circuit breaker state information\n```","ref":"api.html#service-unavailability"},{"type":"extras","title":"Exception Handling - API Specification","doc":"When an exception occurs in the decorated function:\n1. The exception is logged for debugging\n2. A failure is recorded with the exception message as reason\n3. The original exception is re-raised with preserved stacktrace\n\nExample response when circuit is open:\n```elixir\n@decorate Brama.Decorator.circuit_breaker(opts)\ndef my_function(args) do\n  # Function body\nend\n```\n\nWraps functions with circuit breaker functionality.\n\n**Parameters:**\n- `opts` - Options including:\n  - `identifier` - Connection identifier (required)\n  - `error_handler` - Custom function to handle errors and determine success/failure (optional)\n\n**Error Handling:**\nThe error handler can return:\n- `:success` - Operation completed successfully\n- `:failure` - Operation failed\n- `{:failure, reason}` - Operation failed with specific reason\n- `:ignore` - Operation result should not affect circuit state\n\n**Example:**\n```elixir\ndefmodule PaymentService do\n  use Brama.Decorator\n\n  @decorate circuit_breaker(identifier: \"payment_api\")\n  def process_payment(payment) do\n    PaymentAPI.process(payment)\n  end\n  \n  @decorate circuit_breaker(\n    identifier: \"refund_api\",\n    error_handler: fn\n      {:ok, _} -> :success\n      {:error, :invalid_amount} -> {:failure, :validation_error}\n      {:error, :network_timeout} -> {:failure, :service_unavailable}\n      _ -> :ignore\n    end\n  )\n  def process_refund(refund) do\n    RefundAPI.process(refund)\n  end\nend\n```","ref":"api.html#exception-handling"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.3"}}